// This branch history table is a branch taken/not taken predictor
// using a direct mapped cache of two-bit saturating counters.
// The cache memory uses block memory generated by a Xilinx tool named "Core generator".
// It is native to Virtex5, for this to be portable, some ifdefs will need to be implemented.
// The cache is direct mapped using a 10 bit index, and a 20 bit tag.
// Leaving 2 bits in an address for the byte offset, which we assume is always 00
// (we assume word aligned instructions).
//
// The writes are done with port A, and the reads with port B.
// Writes use a 10 bit address, while reads use an 8 bit address.
// The difference in granularity comes from the fact that we predict 4 sequential
// instructions at a time, while we write back to the direct mapped cache only
// 1 prediction at a time.

// Prediction history is stored in 2 bits; along with 20 tag bits that gives us a
// memory usage of 22 bits per predicted branch. With 10 index bits, the memory requres
// (2^10) * 22 bits. In actuality this does not fit perfectly into block memory and we
// get some wasted space. When updating the history bits a saturated increment or decrement
// is done, depending on whether a branch taken or not taken has been observed.
// The most significant bit predicts branch taken or not taken.

module ta_branch_history_table
  (
   input 	i_clk,

   // Read-related wires
   input [31:0] i_address,
   input [31:0] i_address_prev,
   output [3:0] o_branch_taken,
   output [7:0] o_branch_prediction_counters,

   input i_is_subroutine_call_or_return,

   // History update related wires
   input [31:0] i_branch_address, //The address of the branch that was computed
   input 	i_branch_taken_computed, // Whether the branch was computed to be taken or not taken.
   input [1:0] 	i_prediction_history, //The counter value that the branch had when a prediction was made.
   input 	i_branch_valid // Whether the other signals are valid.
   );
`include "ta_functions.v"

   //Important constants.
   // The counter size is 2 bits.
   // The tag size is 20 bits.
   // The index size is 8 bits for reads, and 10 bits for writes.

   /* We need the current PC value (i_address), and the previous PC
    value (i_address_prev). The current PC value is used to read and
    write to memory, while the previous PC value is used to compare
    the tag bits of memory output. In other words, the
    block_memory_output belongs to another clock cycle than i_address.
    */

   // We implement a two-bit saturating up/down computation with a lookup table.
   // I'm sure verilog supports saturating increment/decrement natively somehow,
   // but i could not find the documentation for it.
   wire 	up = i_branch_taken_computed;
   wire [1:0] 	old_counter = i_prediction_history;
   reg [1:0] 	new_counter;
   always @*
     case ({up, old_counter})
       3'b000 : new_counter = 2'b00;
       3'b001 : new_counter = 2'b00;
       3'b010 : new_counter = 2'b01;
       3'b011 : new_counter = 2'b10;
       3'b100 : new_counter = 2'b01;
       3'b101 : new_counter = 2'b10;
       3'b110 : new_counter = 2'b11;
       3'b111 : new_counter = 2'b11;
     endcase

   wire [4 * 22 - 1:0] block_memory_output;

   wire should_update = i_branch_valid && !i_is_subroutine_call_or_return;

   virtex5_branch_history_table
     block_memory
       (
	.clka(i_clk),
	.ena(should_update), //Enable write
	.wea(1'd1), //This is a byte-enable port that we don't use.
	.addra({index(i_branch_address), block_offset(i_branch_address)}),
	.dina({new_counter, tag(i_branch_address)}),
	.clkb(i_clk),
	.addrb(index(i_address)),
	.doutb(block_memory_output)
	);

   genvar 	       i;
   generate
      for(i = 0; i < 4; i = i + 1)
	begin: memory_block
	   wire [21:0] word           = block_memory_output[(i+1)*22-1:i*22];
	   wire [19:0] tag_prediction = word[19:0];
	   wire [1:0]  counter        = word[21:20];
	   wire        hit            = tag(i_address_prev) == tag_prediction;

	   assign o_branch_taken[i]                           = hit & counter[1];
	   assign o_branch_prediction_counters[(i+1) * 2 - 1: i * 2] = hit ? counter : 2'b01;
	end
   endgenerate
endmodule
