// This is a textbook implementation of a BTAC (Branch Target Address Cache,
// also known as a Branch Target Buffer) as described in
// numerous textbooks. It is assumed that the reader understands the workings of a BTAC.

// The main component is a large memory implemented as Xilinx native block memory.
// The memory is dual-ported, with one port dedicated to reads
// and one port dedicated to writes.

/* We need the current PC value (i_address), and the previous PC
 value (address_r). The current PC value is used to read and write
 to memory, while the previous PC value is used to compare the tag
 bits of memory output. In other words, the block_memory_output
 belongs to another clock cycle than i_address.
 */
module ta_branch_target_address_cache
  (
   input 	  i_clk,

   // Read-related wires
   input  [31:0]   i_address,
   input  [31:0]   i_address_prev,
   output [119:0]   o_branch_target,
   output [3:0]    o_cache_hit,

   // Write-related wires
   input 	  i_branch_valid,
   input 	  i_branch_taken_computed,

   // This is the address of a branch instruction that had it's target address computed
   // in decode or execute. It is used to index the BTAC.
   input [19:0]   i_computed_branch_tag,
   input [7:0] 	  i_computed_branch_index,
   input [1:0] 	  i_computed_branch_block_offset,

   // This is an address computed by execute, it is the target of a branch.
   // It is entered into the BTAC so we can "memoize" the branch target.
   input [29:0]   i_branch_target_computed,

   input i_is_subroutine_call_or_return
   );

   // The block memory allows us to enable specific bytes to be written, we do
   // not use this functionality and choose to enable all bytes all the time.
   // Instead we enable writes through the ena input.
   localparam BYTE_ENABLE = 1'd1;

`include "ta_functions.v"

   wire [127:0] block_memory_output;

   wire should_update =  i_branch_valid
                     &&  i_branch_taken_computed
                     && !i_is_subroutine_call_or_return;

   virtex5_block_memory_128x256
     block_memory
       (
	.clka(i_clk),
	.ena (should_update),
	.wea(BYTE_ENABLE),
	.addra({i_computed_branch_index, i_computed_branch_block_offset}),
	.dina({i_branch_target_computed, i_computed_branch_tag[1:0]}),
	.addrb(index(i_address)),
	.doutb(block_memory_output),
	.clkb(i_clk)
	);

   wire [19:0] 	address_tag = tag(i_address_prev);
   genvar 	i;
   generate
      for(i = 0; i < 4; i = i + 1)
	begin: memory_block
	   wire [31:0] 	word = block_memory_output[(i+1) * 32 - 1 : i * 32];
	   // In our BTAC we read 4 words at a time.
	   // A word is 32 bits, the upper 30 bits contain a branch target,
	   // and the lower 2 bits contain a tag.
	   // To get a BTAC hit the 2 tag bits must match the lower two tag bits in
	   // the address_tag.
	   // Note that although an instruction address is 32 bits, the lower 2 bits
	   // are implied to be 00 since instructions are assumed to be word-aligned in memory.
	   assign o_cache_hit[i] = address_tag[1:0] == word[1:0];
	   assign o_branch_target[(i+1)*30-1:i*30] = word[31:2];
	end
   endgenerate
endmodule
