module ta_fetch
  (
   input          i_clk,
   input          i_rst,
   input          i_system_rdy, // External system can stall core with this signal

   // Some commands that come from the coprocessor.
   // They allow the programmer to configure / give commands to the cache.
   input          i_cache_enable, // This comes from the coprocessor.
   input          i_cache_flush, // This comes from the coprocessor.
   input [31:0]   i_cacheable_area, // each bit corresponds to 128MB address space

   // Wishbone
   output         o_wb_req,
   output [31:0]  o_wb_address,
   input [127:0]  i_wb_read_data,
   input          i_wb_ready,

   // Decode and execute
   input [1:0]    i_instruction_ren,
   output [127:0] o_instruction,
   output [1:0]   o_instruction_ready,
   output [63:0]  o_predict_data,
   output         o_is_ras_prediction,

   // Branch computation related wires.
   input          i_branch_valid, //Whether the below signals are valid.
   input [29:0]   i_branch_address,
   input          i_control_hazard,
   input          i_branch_taken_computed,
   input          i_branch_taken_mispredict,
   input [1:0]    i_branch_taken_predicted,
   input [29:0]   i_branch_target_computed,
   input          i_branch_target_mispredict,
   input          i_is_subroutine_call_or_return
   );


wire [31:0]     iaddress;
reg  [31:0]     iaddress_plus0  = 'd0;
wire [31:0]     iaddress_plus4  = iaddress_plus0 + 3'd4;
wire [31:0]     iaddress_plus8  = iaddress_plus0 + 4'd8;
wire [31:0]     iaddress_plus12 = iaddress_plus0 + 4'd12;

// Cache signals
wire         cache_hit;
wire         cache_busy;
wire [127:0] cache_read_data128;
wire         address_cachable   = i_cacheable_area[iaddress_plus0[31:27]];
wire         sel_cache          = address_cachable && i_cache_enable;
wire         cache_miss         = !cache_busy && !cache_hit;

// Misc
wire   fetch_stalling =      !i_system_rdy
                           || o_wb_req
                           || cache_busy;

reg wb_req_r = 'd0; // Holds wb request signal until wb data is ready
wire        pc_selector_flush;
wire        fifo_overflow;
wire [3:0]  btac_cache_hit;
wire [3:0]  bht_branch_taken;
wire [7:0]  bht_branch_counter;
wire [3:0]  selector_selected;
wire [29:0] ras_branch_target;
wire [3:0]  ras_valid_branch;
wire        any_ras_valid_branch = |(ras_valid_branch);
wire [3:0]  fifo_wen = {4{!fetch_stalling}} & selector_selected ;

// Various modules are sensitive to different signals wrt. stalling
wire        ras_stall = fetch_stalling || pc_selector_flush || fifo_overflow;
wire        pc_selector_stall = fetch_stalling||fifo_overflow;

// Instructions predicted as taken, either by the BHT or the RAS
wire [3:0]  instruction_selector_branch_taken = (bht_branch_taken & btac_cache_hit) | ras_valid_branch;

// Which of the four instructions that is predicted as a taken branch.
wire [3:0]  predicted_branch_taken = selector_selected & bht_branch_taken;

wire [119:0] btac_predicted_branch_target; // Written to by BTAC
wire [29:0]  chosen_btac_branch_target = predicted_branch_taken[0] ? btac_predicted_branch_target[29:0]  :
                                         predicted_branch_taken[1] ? btac_predicted_branch_target[59:30] :
                                         predicted_branch_taken[2] ? btac_predicted_branch_target[89:60] :
                                                                     btac_predicted_branch_target[119:90];

wire [7:0] new_predicted_branch_counter = bht_branch_counter |
                                         {
                                         {2{ras_valid_branch[3]}},
                                         {2{ras_valid_branch[2]}},
                                         {2{ras_valid_branch[1]}},
                                         {2{ras_valid_branch[0]}}
                                         };

// Both RAS and BTAC can predict a branch
wire [29:0] new_predicted_branch_target = any_ras_valid_branch       ?
                                          ras_branch_target          :
                                          chosen_btac_branch_target  ;

// Only allow a wb request from one source at a time
wire wb_req = sel_cache? cache_miss : !fifo_overflow;

// Outputs
assign o_wb_address = iaddress;
assign o_wb_req     = (wb_req || wb_req_r) && !i_wb_ready;

// Return read data either from the wishbone bus or the cache
// If we miss in the cache we read directly from wb when the data is ready,
// omitting the cache write turn around time
wire [127:0] selected_read_data128 = (sel_cache && cache_hit) ? cache_read_data128 : i_wb_read_data;

always @(posedge i_clk)
  begin
     wb_req_r         <= i_rst ?  1'd0 : (wb_req_r || wb_req) && !i_wb_ready;
     iaddress_plus0   <= i_rst ? 32'd0 : {iaddress[31:4], 4'd0};
  end

// Debug purposes... can be removed anytime.
  reg [31:0] ras_misprediction = 'd0;
always @ ( posedge i_clk )
    if (i_branch_valid && i_branch_target_mispredict && i_is_subroutine_call_or_return)
      ras_misprediction <= ras_misprediction + 1'd1;


   // ======================================
   // L1 Instruction Cache
   // ======================================
   ta_icache
      u_cache
       (
    .i_clk                      ( i_clk                               ),
    .i_rst                      ( i_rst                               ),
    .o_busy                     ( cache_busy                          ),
    .o_hit                      ( cache_hit                           ),
    .i_enable                   ( sel_cache                           ),
    .i_address                  ( iaddress                            ),
    .i_cache_flush              ( i_cache_flush                       ),
    .o_read_data                ( cache_read_data128                  ),
    .i_read_data                ( i_wb_read_data                      ),
    .i_data_ready               ( i_wb_ready                          )
    );

   // ======================================
   // Branch History Table
   // ======================================
   ta_branch_history_table
     u_branch_history_table
       (
    .i_clk                           ( i_clk                          ),
    .i_address                       ( iaddress                       ),
    .i_address_prev                  ( iaddress_plus0                 ),
    .o_branch_taken                  ( bht_branch_taken               ),
    .o_branch_prediction_counters    ( bht_branch_counter             ),
    .i_branch_address                ( {i_branch_address, 2'b00}      ),
    .i_branch_taken_computed         ( i_branch_taken_computed        ),
    .i_is_subroutine_call_or_return  ( i_is_subroutine_call_or_return ),
    .i_prediction_history            ( i_branch_taken_predicted       ),
    .i_branch_valid                  ( i_branch_valid                 )
    );

   // ======================================
   // Branch Target Address Cache
   // ======================================
   ta_branch_target_address_cache
     u_branch_target_address_cache
       (
    .i_clk                           ( i_clk                         ),
    .i_address                       ( iaddress                      ),
    .i_address_prev                  ( iaddress_plus0                ),
    .o_branch_target                 ( btac_predicted_branch_target  ),
    .o_cache_hit                     ( btac_cache_hit                ),
    .i_branch_valid                  ( i_branch_valid                ),
    .i_branch_taken_computed         ( i_branch_taken_computed       ),
    .i_is_subroutine_call_or_return  ( i_is_subroutine_call_or_return),
    .i_computed_branch_tag           ( i_branch_address[29:10]       ),
    .i_computed_branch_index         ( i_branch_address[9:2]         ),
    .i_computed_branch_block_offset  ( i_branch_address[1:0]         ),
    .i_branch_target_computed        ( i_branch_target_computed      )
    );

   // ======================================
   // Address Generation Unit
   // ======================================
   ta_pc_selector
     u_pc_selector
       (
    .i_clk                       ( i_clk                         ),
    .i_rst                       ( i_rst                         ),
    .i_btac_mispredict           ( i_branch_target_mispredict    ),
    .i_btac_cache_hit            ( btac_cache_hit                ),
    .i_btac_branch_target        ( chosen_btac_branch_target     ),
    .i_bht_mispredict            ( i_branch_taken_mispredict     ),
    .i_predicted_branch_taken    ( predicted_branch_taken        ),
    .i_ras_valid_branch          ( any_ras_valid_branch          ),
    .i_ras_branch_target         ( ras_branch_target             ),
    .i_branch_valid              ( i_branch_valid                ),
    .i_stall                     ( pc_selector_stall             ),
    .i_exe_cmp_branch_target     ( i_branch_target_computed      ),
    .i_exe_branch_address        ( i_branch_address              ),
    .i_exe_cmp_branch_taken      ( i_branch_taken_computed       ),
    .i_wb_req_r                  ( wb_req_r                      ),
    .i_control_hazard            ( i_control_hazard              ),
    .o_pc                        ( iaddress                      ),
    .o_flush                     ( pc_selector_flush             )
    );

   // ======================================
   // Instruction Selector
   // ======================================
   ta_instruction_selector
     u_instruction_selector
    (
    .i_clk               ( i_clk                             ),
    .i_block_offset      ( iaddress[3:2]                     ),
    .i_p_branch_taken    ( instruction_selector_branch_taken ),
    .o_selected          ( selector_selected                 )
    );

   // ======================================
   // Instruction queue
   // ======================================
   ta_fifo_4way_write_2way_read
     u_fifo
    (
    .i_clk               ( i_clk                                                            ),
    .i_rst               ( i_rst || pc_selector_flush                                       ),
    .i_wen               ( fifo_wen                                                         ),
    .i_ren               ( i_instruction_ren                                                ),
    .i_pc                ( {iaddress_plus12, iaddress_plus8, iaddress_plus4, iaddress_plus0}),
    .i_instruction       ( selected_read_data128                                            ),
    .i_predict_target    ( new_predicted_branch_target                                      ),
    .i_predict_counter   ( new_predicted_branch_counter                                     ),
    .i_is_ras_prediction ( ras_valid_branch                                                 ),
    .o_instruction       ( o_instruction                                                    ),
    .o_predict_data      ( o_predict_data                                                   ),
    .o_ready             ( o_instruction_ready                                              ),
    .o_overflow          ( fifo_overflow                                                    ),
    .o_is_ras_prediction (o_is_ras_prediction                                            )
    );

   // ======================================
   // Return Address Stack
   // ======================================
   ta_return_address_stack
     u_return_address_stack
       (
    .i_clk                      ( i_clk                     ),
    .i_rst                      ( i_rst                     ),
    .i_instruction              ( selected_read_data128     ),
    .i_stall                    ( ras_stall                 ),
    .i_valid_instr              ( selector_selected         ),
    .i_pc                       ( {iaddress_plus12, iaddress_plus8, iaddress_plus4, iaddress_plus0} ),
    .o_valid_branch             ( ras_valid_branch          ),
    .o_branch_target            ( ras_branch_target         )
   );
 
endmodule
