/*
 FIFO implemented by using a circular buffer.
 ----------------
 lower half of o_instruction always output the instruction which
 should be executed first.
*/

module ta_fifo_4way_write_2way_read(
  input                 i_clk,
  input                 i_rst,
  input [3:0]           i_wen,
  input [1:0]           i_ren,
  input [32 * 4 - 1:0]  i_pc,
  input [32 * 4 - 1:0]  i_instruction,
  input [29:0]          i_predict_target,
  input [7:0]           i_predict_counter,
  input [3:0]           i_is_ras_prediction,

  output                o_overflow,
  output [1:0]          o_ready,
  output [127:0]        o_instruction,
  output [63:0]         o_predict_data,
  output                o_is_ras_prediction
);

`include "ta_functions.v"
// Must be a power of 2
localparam FIFO_ENTRIES = 8;
localparam LOG2_FIFO_ENTRIES = log2(FIFO_ENTRIES); //Must be power of 2.
// Data     32 bit
// PC       32 bit
// Predict  32 bit
// Ras pred  1 bit
// ---------------
//          97 bit
localparam LINE_WIDTH = 97;

reg [LOG2_FIFO_ENTRIES - 1:0] head  = 'd0;
reg [LOG2_FIFO_ENTRIES - 1:0] tail  = 'd0;
reg [LOG2_FIFO_ENTRIES    :0] count = 'd0;
   
assign o_ready = {|count[LOG2_FIFO_ENTRIES:1], |count};

/* Number of valid reads on next cycle */
wire [1:0] nof_v_reads = (i_ren[1] && o_ready[1]) + (i_ren[0] && o_ready[0]);

/* Number of potential writes on next cycle */
wire [2:0] nof_writes = i_wen[0] + i_wen[1] + i_wen[2] + i_wen[3];


/* Number of acutal writes on next cycle */
wire [2:0] nof_v_writes = o_overflow ? 0 : nof_writes;

wire [2:0] head_nxt = head + nof_v_reads;
wire [2:0] tail_nxt = tail + nof_v_writes;
wire [3:0] count_nxt = count + nof_v_writes - nof_v_reads;

reg [LINE_WIDTH-1:0] storage [FIFO_ENTRIES-1:0];


always @( posedge i_clk )
  if (!o_overflow)
    begin
    if (i_wen[0])
      storage[tail] =
                        { i_is_ras_prediction[0], i_predict_target, i_predict_counter[1:0] , i_pc[31:0], i_instruction[31:0] };
    if (i_wen[1])
      storage[tail + i_wen[0]] =
                        { i_is_ras_prediction[1], i_predict_target, i_predict_counter[3:2] , i_pc[63:32], i_instruction[63:32] };
    if (i_wen[2])
      storage[tail + i_wen[0] + i_wen[1]] =
                        { i_is_ras_prediction[2], i_predict_target, i_predict_counter[5:4] , i_pc[95:64], i_instruction[95:64] };
    if (i_wen[3])
      storage[tail + i_wen[0] + i_wen[1] + i_wen[2]] =
                        { i_is_ras_prediction[3], i_predict_target, i_predict_counter[7:6] , i_pc[127:96], i_instruction[127:96] };
    end

always @( posedge i_clk )
  begin
    if (i_rst)
      begin
      tail = 3'd0;
      count = 3'd0;
      head = 4'd0;
      end
    else
      begin
      tail = tail_nxt;
      count = count_nxt;
      head = head_nxt;
      end
end

assign o_overflow = count > FIFO_ENTRIES - 4; //count == FIFO_ENTRIES ||
                    //count //+ nof_writes - nof_v_reads > FIFO_ENTRIES;

assign o_instruction  = count == 4'd0 ?
                        128'he1a00000 : 
                        { storage[head + 1'b1][63:0] , storage[head][63:0]  };

assign o_is_ras_prediction = storage[head][96];

assign o_predict_data = count == 4'd0 ?
                        64'd0 : { storage[head + 1'b1][95:64], storage[head][95:64] };

endmodule
