module ta_pc_selector (
    input         i_clk,
    input         i_rst,

    // BTAC
    input         i_btac_mispredict,
    input [3:0]   i_btac_cache_hit,
    input [29:0]  i_btac_branch_target,

    // Branch predictor
    input         i_bht_mispredict,

    input [3:0]    i_predicted_branch_taken,
    // Return Address Stack
    input         i_ras_valid_branch,
    input [29:0]  i_ras_branch_target,

    input         i_branch_valid,

    // Cache
    input         i_stall,
    // Execute
    input [29:0]  i_exe_cmp_branch_target,
    input [29:0]  i_exe_branch_address,
    input         i_exe_cmp_branch_taken,
    // Wishbone
    input         i_wb_req_r,

    input         i_control_hazard,
    output [31:0] o_pc,
    output        o_flush
);

reg  [29:0] pc_mispredict_corrected;
reg  [29:0] pc_prev= 'd0;
wire [29:0] pc_nxt = { pc_prev[29:2] + 1'b1, 2'd0 };

localparam NORMAL  = 1'd0,
           WAIT_WB = 1'd1;

reg control_state     = 'd0;
reg control_state_nxt = 'd0;

always @( posedge i_clk)
    control_state = control_state_nxt;


/*
 * This state machine ignores any pending wishbone request if we get a control hazard during a wishbone access
 */
always @*
    case (control_state)
        NORMAL:     control_state_nxt = (i_control_hazard && i_wb_req_r) ? WAIT_WB : NORMAL;
        WAIT_WB:    control_state_nxt =                      i_wb_req_r  ? WAIT_WB : NORMAL;
    endcase

/* Sometimes the branch predictor will hit, but the BTAC will miss.
 * Then we assume not taken, since we don't know the target address */
wire valid_predicted_branch = |(i_predicted_branch_taken & i_btac_cache_hit);

/*
 *   mispredict  |
 * branch | btac | cmp_brch_tkn  | Correct PC
 *   0    |   0  |      0        | OK
 *   0    |   0  |      1        | OK
 *   0    |   1  |      0        | OK
 *   0    |   1  |      1        | PC = CMP_BRANCH_TARGET
 *   1    |   0  |      0        | PC = BRANCH_ADDRESS + 1
 *   1    |   0  |      1        | PC = CMP_BRANCH_TARGET
 *   1    |   1  |      0        | PC = BRANCH_ADDRESS + 1
 *   1    |   1  |      1        | PC = CMP_BRANCH_TARGET
 */

always @*
casez( {i_bht_mispredict, i_btac_mispredict, i_exe_cmp_branch_taken} )
    3'b011:     pc_mispredict_corrected = i_exe_cmp_branch_target;
    3'b1?0:     pc_mispredict_corrected = i_exe_branch_address + 1'b1;
    3'b1?1:     pc_mispredict_corrected = i_exe_cmp_branch_target;
    default:    pc_mispredict_corrected = 30'dx;
endcase

/* Order is important */
wire[29:0]  pc = i_rst                         ?   0                       :
                 i_control_hazard              ?   pc_mispredict_corrected :
                 i_stall
                 || (control_state == WAIT_WB) ?   pc_prev                 :
                 i_ras_valid_branch            ?   i_ras_branch_target     :
                 valid_predicted_branch        ?   i_btac_branch_target    :
                                                   pc_nxt                  ;
always @( posedge i_clk)
    pc_prev = pc;

assign o_pc = {pc, 2'd0};
assign o_flush =    i_bht_mispredict
                || (i_btac_mispredict && i_exe_cmp_branch_taken)
                || (control_state == WAIT_WB);

endmodule
