/**
  * Return Address Stack
 **/

module ta_return_address_stack (
  input          i_clk,
  input          i_rst,
  input          i_stall,
  input  [127:0] i_instruction,
  input  [3:0]   i_valid_instr,
  input  [127:0] i_pc,
  output [3:0]   o_valid_branch,
  output [29:0]  o_branch_target
);

`include "ta_functions.v"

localparam ENTRIES = 'd16;
localparam LOG2_ENTRIES = log2(ENTRIES);

// Next sequential address (30 bit)
localparam WIDTH = 'd30;

reg [WIDTH-1:0] storage [ENTRIES-1:0];
reg [LOG2_ENTRIES - 1:0] top = 'd0;

wire [3:0] is_call_instr;
wire [3:0] is_ret_instr;

integer j;
initial
    for (j = 0; j < ENTRIES; j = j + 1)
        storage[j] <= 30'd0;

genvar i;
generate
  for (i = 0; i < 4; i = i + 1)
    begin: o_valid_branch_generate
      assign is_call_instr[i] = i_valid_instr[i] && i_instruction[ (i+1)*32 - 1 : i*32 + 24] == 8'b11101011;
      assign is_ret_instr[i] =  i_valid_instr[i] &&
                               (
                                 i_instruction[ (i+1)*32 - 1 : i*32] == MOV_PC_LR   ||
                                                        // We don't care about the lower 15 bits
                                 i_instruction[ (i+1)*32 - 1 : i*32 + 15] == POP_PC
                                ) ;
    end
endgenerate

assign o_valid_branch = is_call_instr | is_ret_instr;

wire [29:0] return_target = storage[top] + 30'd1;

wire [31:0] decode_target = is_call_instr[0] ? i_instruction[31:0]  :
                            is_call_instr[1] ? i_instruction[63:32] :
                            is_call_instr[2] ? i_instruction[95:64] :
                                               i_instruction[127:96];

wire [29:0] decode_pc     = o_valid_branch[0] ? i_pc[31:2]  :
                            o_valid_branch[1] ? i_pc[63:34] :
                            o_valid_branch[2] ? i_pc[95:66] :
                                                i_pc[127:98];

wire [29:0] decoded_target = {{6{decode_target[23]}}, decode_target[23:0]} + 4'd2 + decode_pc;

wire [1:0] ras_index = o_valid_branch[0] ? 0 :
                       o_valid_branch[1] ? 1 :
                       o_valid_branch[2] ? 2 :
                                           3 ;

assign o_branch_target = is_call_instr[ras_index] ? decoded_target : return_target;

always @ (posedge i_clk)
    if (!(i_rst || i_stall))
    begin
      if (is_call_instr[ras_index])
      begin
        top = top + 1'd1;
        storage[top] = decode_pc;
      end
      else if (is_ret_instr[ras_index])
        top = top - 1'd1;
    end

endmodule
